#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This program is free software. It comes without any warranty, to the extent
permitted by applicable law. You can redistribute it and/or modify it under the
terms of the Do What The Fuck You Want To Public License, Version 2, as
published by Sam Hocevar. See http://www.wtfpl.net/ for more details.
"""

import argparse
import json
import os
import re
import shutil
import time

import requests
import lxml.html

BASEURL = 'https://studip.uni-hannover.de/api.php'
LOGIN_URL = "https://studip.uni-hannover.de/Shibboleth.sso/Login?target=https%3A%2F%2Fstudip.uni-hannover.de%2Findex.php%3Fagain%3Dyes%26sso%3Dshib"
SAML_URL = "https://studip.uni-hannover.de/Shibboleth.sso/SAML2/POST"

#TODO: mayyyybe do all this as an object?

def main(argv=None):

    user, password, path, allsemesters, blacklist, ignore = parse_args()
    cookie = sso_login(user, password)
    with requests.session() as session:
        session.cookies = requests.utils.cookiejar_from_dict(cookie)
        scrape(session, path, allsemesters, blacklist, ignore)

def parse_args():
    try:
        import keyring
    except ImportError:
        usekeyring = False
    else:
        usekeyring = True

    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--path', help='Pfad für lokalen Dateibaum')
    parser.add_argument('-u', '--user', help='Stud.IP Benutzername')
    parser.add_argument('-pw', '--password', help='Stud.IP Passwort')
    parser.add_argument('-b', '--blacklist', help='Blacklist für Veranstaltungen, mehrmals angeben für mehrere Veranstaltungen', action='append')
    parser.add_argument('-i', '--ignore', help='Blacklist für Dateinamen als regex')
    parser.add_argument('-a', '--allsemesters', help='Alle Semester statt nur dem aktuellen', action="store_true")
    parser.add_argument('-c', '--config', help='JSON-Konfigurationsdatei, wird ohne Angabe in $XDG_CONFIG_HOME/studsauger/config.json oder $HOME/.config/studsauger/config.json gesucht')
    args = parser.parse_args()

    configpath = None
    if args.config: 
        if os.path.isfile(args.config):
            configpath = args.config
        else:
            print('Konfigurationsdatei "' + args.config + '" existiert nicht.')
            exit(1)
    elif 'XDG_CONFIG_HOME' in os.environ:
        if os.path.isfile(os.environ['XDG_CONFIG_HOME'] + '/studsauger/config.json'):
            configpath = os.environ['XDG_CONFIG_HOME'] + '/studsauger/config.json'
    elif 'HOME' in os.environ:
        if os.path.isfile(os.environ['HOME'] + '/.config/studsauger/config.json'):
            configpath = os.environ['HOME'] + '/.config/studsauger/config.json'

    if configpath:
        with open(configpath, 'r') as stream:
            config = json.load(stream)
    else: 
        config = {}

    path = None
    if args.path:
        path = args.path
    elif 'path' in config:
        path = config['path']
    else:
        print('Kein Pfad angegeben!')
        parser.print_help()
        exit(1)

    if args.user:
        user = args.user
    elif 'user' in config:
        user = config['user']
    else:
        print('Kein Benutzer angegeben!')
        parser.print_help()
        exit(1)

    if usekeyring:
        keyringpw = keyring.get_password('studsauger', user)
        if keyringpw:
            password = keyringpw
            if args.password or 'password' in config:
                print('Passwort für Benutzer "' + user + '" ist im Keyring und darf nicht angegeben werden!')
            exit(1)
    elif args.password:
        password = args.password
    elif 'password' in config:
        password = config['password']
    else:
        print('Kein Passwort angegeben!')
        exit(1)

    if args.blacklist:
        blacklist = args.blacklist
    elif 'blacklist' in config:
        blacklist = config['blacklist']
    else:
        blacklist = {}
    if args.ignore:
        ignore = args.ignore
    elif 'ignore' in config:
        ignore = config['ignore']
    else:
        ignore = None
    if args.allsemesters:
        allsemesters = True
    elif 'allsemesters' in config:
        allsemesters = config['allsemesters']
    else:
        allsemesters = False

    return user, password, path, allsemesters, blacklist, ignore


#TODO: return session instead pf cookie for max aesthetics
# beautiful POS, returns the Seminar_Session Cookie necessary for accessing the API
def sso_login(username, password):
    # Start session and get login form.
    with requests.session() as session:
        session.auth = (username, password)
        login = session.get(LOGIN_URL)
        form = {'j_username': username, 'j_password': password}
        response = session.post(login.url, data=form)

        SAML_html = lxml.html.fromstring(response.text)
        hidden_inputs = SAML_html.xpath(r'//form//input[@type="hidden"]')
        form2 = {x.attrib["name"]: x.attrib["value"] for x in hidden_inputs}
        response2 = session.post(SAML_URL, data=form2)
        return dict(Seminar_Session=session.cookies['Seminar_Session'])

def scrape(session, path, allsemesters, blacklist, ignore):
    json_user = session.get(BASEURL + '/user').json()
    json_semesters = session.get(BASEURL + '/semesters', params='limit=1000').json()
    json_courses = session.get(BASEURL + '/user/' + json_user['user_id'] + '/courses', params='limit=1000').json()

    if not allsemesters:
        latest = None
        unixnow = int(time.time())
        for key, semester in json_semesters['collection'].items():
            if unixnow >= semester['begin'] and unixnow <= semester['end']:
                latest = semester['id']
                json_courses = session.get(BASEURL + '/user/' + json_user['user_id'] + '/courses?semester=' + latest, params='limit=1000').json()
                break
        if not latest:
            print('Aktuelles Semester nicht gefunden')
            exit(1)

    if os.path.isfile(path + '/.database.json'):
        with open(path + '/.database.json', 'r') as databasefile:
            database = json.load(databasefile)
    else:
        database = {}

    for key, course in json_courses['collection'].items():
        coursename = course['title']
        if coursename in blacklist:
            continue
        print('=== ' + coursename + ' ===')
        mkdir(path + '/' + coursename.replace('/', '_'))
        json_files = session.get(BASEURL + '/course/' + course['course_id'] + '/top_folder', params='limit=1000').json()
        traverseFileTree(json_files, path, ignore, database, session, coursename)
    with open(path + '/.database.json', 'w') as databasefile:
        json.dump(database, databasefile, indent=4)

def traverseFileTree(root, path, ignore, database, session, coursename):
    ### download all files in current folders
    if 'file_refs' in root:
        for fileref in root['file_refs']:
            if ignore and re.match(ignore, fileref['name']):
                continue
            date = ''
            if fileref['file_id'] in database:
                if database[fileref['file_id']] == fileref['chdate']:
                    continue
                else:
                    date = fileref['chdate']
            database[fileref['file_id']] = fileref['chdate']
            print(fileref['name'] + str(date))
            target = path + '/' + coursename.replace('/', '_') + '/' + fileref['name'] + str(date)
            download = session.get(BASEURL + '/file/' + fileref['id'] + '/download', stream=True)
            with open(target, 'wb') as downloadfile:
                download.raw.decode_content = True
                shutil.copyfileobj(download.raw, downloadfile)
    ### Look for subfolders
    if 'subfolders' in root and root['subfolders']:
        for json_subfolder in root['subfolders']:
            subfolder = session.get(BASEURL + '/folder/' + json_subfolder['id'], params='limit=1000').json()
            #TODO: implement making subfolders, for people that really like clicking a lot?????
            #subpath = path + '/' + coursename.replace('/', '_')
            #mkdir(subpath)
            traverseFileTree(subfolder, path, ignore, database, session, coursename)


def mkdir(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)

main()
